API Exits - a new interface

Mark E Taylor
Technical Strategist, IBM Hursley (UK)	c IBM 2002


     setmqaut -t q -m qmgr -n APIX.STAT -g nobody +put


/* MODULE: apixstat.c
 * DESCRIPTION: Example API Exit to count MQ message operations
 * AUTHOR: Mark Taylor, IBM Hursley
 * DATE: January 2002
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmqc.h>
#include <cmqxc.h>
/* Definition of what we're going to log. This contains info about 
 * all the messages put and got by this application, no matter what 
 * queue. I could start tracking number of open queues, operations per   
 * queue etc, but that starts to make this example too big and would 
 * obscure the basic exit.
 */
struct mqstats {
    MQLONG Environment;
    long BytesPut;
    long BytesGot;
    long BytesBrowsed;
    long MsgPut;
    long MsgGot;
    long MsgBrowsed;
    long MaxLengthPut;
    long MaxLengthGot;
};
/* Standard MQ Entrypoint. Not directly used, but
 * required by some platforms.
 */
void MQStart(){;}
/* Declare internal functions. All except
 * the entrypoint can be static as they're not used externally.
 */
MQ_INIT_EXIT EntryPoint;
static MQ_TERM_EXIT Terminate;
static MQ_PUT_EXIT  PutCounter;
static MQ_PUT1_EXIT Put1Counter;
static MQ_GET_EXIT  GetCounter;
static MQ_DISC_EXIT WriteStats;
static void ClientCheck(struct mqstats *,PMQAXP,PMQAXC);
/* Collect statistics on messages put to queues.
 * Need to do this for both MQPUT and MQPUT1 calls.
 */
static void MQENTRY PutCounter ( PMQAXP    pExitParms
                    , PMQAXC    pExitContext
                    , PMQHCONN  pHconn
                    , PMQHOBJ   pHobj
                    , PPMQMD    ppMsgDesc
                    , PPMQPMO   ppPutMsgOpts
                    , PMQLONG   pBufferLength
                    , PPMQVOID  ppBuffer
                    , PMQLONG   pCompCode
                    , PMQLONG   pReason
                    )
{
  struct mqstats **ppmqstats = (void *)&pExitParms->ExitUserArea;
  struct mqstats *pmqstats = *ppmqstats;
  if (pmqstats->Environment == -1)
    ClientCheck(pmqstats,pExitParms,pExitContext);
  pmqstats->MsgPut ++;
  pmqstats->BytesPut += *pBufferLength;
  if (*pBufferLength > pmqstats->MaxLengthPut)
    pmqstats->MaxLengthPut = *pBufferLength;
  return;
}
static void MQENTRY Put1Counter ( PMQAXP    pExitParms
                    , PMQAXC    pExitContext
                    , PMQHCONN  pHconn
                    , PPMQOD    ppObjDesc
                    , PPMQMD    ppMsgDesc
                    , PPMQPMO   ppPutMsgOpts
                    , PMQLONG   pBufferLength
                    , PPMQVOID  ppBuffer
                    , PMQLONG   pCompCode
                    , PMQLONG   pReason
                    )
{
  struct mqstats **ppmqstats = (void *)&pExitParms->ExitUserArea;
  struct mqstats *pmqstats = *ppmqstats;
  if (pmqstats->Environment == -1)
    ClientCheck(pmqstats,pExitParms,pExitContext);
  pmqstats->MsgPut ++;
  pmqstats->BytesPut += *pBufferLength;
  if (*pBufferLength > pmqstats->MaxLengthPut)
    pmqstats->MaxLengthPut = *pBufferLength;
  return;
}
/* Collect statistics on messages got from queues.
 * Only update the stats if the retrieval was successful (or there
 * was a warning, probably due to truncation). Browsed messages
 * are monitored separately from destructive retrieval.
 */
static void MQENTRY GetCounter ( PMQAXP    pExitParms
                    , PMQAXC    pExitContext
                    , PMQHCONN  pHconn
                    , PMQHOBJ   pHobj
                    , PPMQMD    ppMsgDesc
                    , PPMQGMO   ppGetMsgOpts
                    , PMQLONG   pBufferLength
                    , PPMQVOID  ppBuffer
                    , PPMQLONG  ppDataLength
                    , PMQLONG   pCompCode
                    , PMQLONG   pReason
                    )
{
  MQLONG options;
  MQLONG browseopts = (MQGMO_BROWSE_FIRST |
                  MQGMO_BROWSE_NEXT  |
                  MQGMO_BROWSE_MSG_UNDER_CURSOR);
  PMQGMO gmo = *ppGetMsgOpts;
  struct mqstats **ppmqstats = (void *)&pExitParms->ExitUserArea;
  struct mqstats *pmqstats = *ppmqstats;
  if (pmqstats->Environment == -1)
    ClientCheck(pmqstats,pExitParms,pExitContext);
  if (*pCompCode != MQCC_FAILED )
  {
    options = gmo->Options;
    if (options & browseopts)
    {
      pmqstats->MsgBrowsed ++;
      pmqstats->BytesBrowsed += **ppDataLength;
    }
    else
    {
      pmqstats->MsgGot ++;
      pmqstats->BytesGot += **ppDataLength;
      if  (**ppDataLength > pmqstats->MaxLengthGot)
   pmqstats->MaxLengthGot = **ppDataLength;
    }
  }
  return;
}
/* This function is called during MQDISC, before the actual disconnect
 * has happened.It builds a message from the statistics gathered and
 * puts it to a queue, which must have been previously defined.
 * The defaults are used for putting messages, except for forcing an
 * out-of-sync operation.
 * This MQPUT1 does not itself get passed to an API exit, as the calls
 * are not recursive.
 */
static void MQENTRY WriteStats ( PMQAXP    pExitParms
                    , PMQAXC    pExitContext
                    , PPMQHCONN ppHconn
                    , PMQLONG   pCompCode
                    , PMQLONG   pReason
                    )
{
  MQLONG LocalCompCode = MQCC_OK;
  MQLONG LocalReason   = MQRC_NONE;
  MQLONG BufferLength;
  MQBYTE Buffer[1024];
  MQMD   MsgDesc = {MQMD_DEFAULT};
  MQOD   ObjDesc = {MQOD_DEFAULT};
  MQPMO  PMO     = {MQPMO_DEFAULT};
  struct mqstats **ppmqstats = (void *)&pExitParms->ExitUserArea;
  struct mqstats *pmqstats = *ppmqstats;
  sprintf((char *)Buffer,
  "App:%-28.28s User:%-12.12s Env:%08X MsgPut:%d MsgGot:%d MsgBrw:%d BytesPut:%d BytesGot:%d BytesBrw: %d MaxPut:%d MaxGot:%d\n",
    (pmqstats->Environment == MQXE_MCA_SVRCONN)?
   "ClientProxy" : pExitContext->ApplName,
    pExitContext->UserId,
    pmqstats->Environment,
    pmqstats->MsgPut,pmqstats->MsgGot,pmqstats->MsgBrowsed,
    pmqstats->BytesPut,pmqstats->BytesGot,pmqstats->BytesBrowsed,
    pmqstats->MaxLengthPut,pmqstats->MaxLengthGot);
  BufferLength = strlen((char *)Buffer);
  strcpy(ObjDesc.ObjectName,"APIX.STAT");
  memcpy(MsgDesc.Format,MQFMT_STRING,8);
  PMO.Options |= MQPMO_NO_SYNCPOINT;
  PMO.Options |= MQPMO_FAIL_IF_QUIESCING;
  /* Don't use the application's CC/RC variables, as
   * we want to continue even if this MQPUT1 fails.
   */
  MQPUT1(**ppHconn,
    &ObjDesc,
    &MsgDesc,
    &PMO,
    BufferLength,
    Buffer,
    &LocalCompCode,
    &LocalReason
    );
  return;
}
/* This is called just after the MQDISC, and can be used to
 * tidy-up any resources we've used during the run.
 */
static void MQENTRY Terminate ( PMQAXP   pExitParms
                  , PMQAXC   pExitContext
                  , PMQLONG  pCompCode
                  , PMQLONG  pReason
                  )
{
  struct mqstats **ppmqstats = (void *)&pExitParms->ExitUserArea;
  struct mqstats *pmqstats = *ppmqstats;
  free(pmqstats);
  return;
}
/* Initialization function. This will be called each time an
 * application connects to the queue manager.
 */
void MQENTRY EntryPoint ( PMQAXP   pExitParms
                   , PMQAXC   pExitContext
                   , PMQLONG  pCompCode
                   , PMQLONG  pReason
                   )
{
  MQLONG            rc             = MQRC_NONE;
  char buf[256]={0};
  struct mqstats **ppmqstats = (void *)&pExitParms->ExitUserArea;
  struct mqstats *pmqstats = NULL;
  /* This version string will be logged in MQ's trace so we
   * can see it's been called.
   */
#ifdef _REENTRANT
  char *t="threaded";
#else
  char *t="unthreaded";
#endif
  sprintf(buf,"APIXSTAT ver %s %s (%s)\n",__DATE__,__TIME__,t);
  memcpy(pExitParms->ExitPDArea,buf,48);
  /* This is the anchor point for our per-connect information.
   * The pointer can be stored in the ExitParms structure, and will be
   * returned on all future calls by this particular thread/connection.
   * We don't need to worry about threads or locks when updating the
   * MQAXP structure as thread-specific versions have already been
   * allocated by the queue manager. Of course, any system calls made
   * here (like malloc) will need to be thread-safe.
   * ExitUserArea is a 16 byte array, easily large enough to hold a 
   * pointer value.
   */
  pmqstats = malloc(sizeof(struct mqstats));
  if (pmqstats)
  {
    *ppmqstats = pmqstats;
    bzero(pmqstats,sizeof(struct mqstats));
    pmqstats->Environment = pExitContext->Environment;
    if (pmqstats->Environment == MQXE_MCA)
      pmqstats->Environment = -1; /* Mark as unknown for now ... */
  }
  else
  {
    /* If the malloc failed, then fail as fast as possible.
     */
    rc = MQRC_API_EXIT_ERROR;
    pExitParms->ExitResponse = MQXCC_FAILED;
  }
/* Register for the verbs we're interested in - MQPUT(1), MQGET, MQDISC
* and for the terminate routine so we can clear up allocated resources.
* If I wanted to track only committed message operations then I'd need 
* to also intercept MQCMIT and MQBACK, and keep counts for the state of
* PUT and GOT messages. That's too much work for this example.
   */
  if (rc == MQRC_NONE)
  {
    MQXEP ( pExitParms->Hconfig, MQXR_BEFORE, MQXF_PUT
     , (PMQFUNC) PutCounter, 0, pCompCode, pReason);
    MQXEP ( pExitParms->Hconfig, MQXR_BEFORE, MQXF_PUT1
     , (PMQFUNC) PutCounter, 0, pCompCode, pReason);
    MQXEP ( pExitParms->Hconfig , MQXR_AFTER, MQXF_GET
     , (PMQFUNC) GetCounter, 0, pCompCode, pReason);
    MQXEP ( pExitParms->Hconfig , MQXR_BEFORE, MQXF_DISC
     , (PMQFUNC) WriteStats, 0, pCompCode, pReason);
    MQXEP ( pExitParms->Hconfig , MQXR_CONNECTION, MQXF_TERM
     , (PMQFUNC) Terminate,  0, pCompCode, pReason);
  }
  if (rc != MQRC_NONE)
  {
    pExitParms->ExitResponse = MQXCC_FAILED;
  }
  return;
}
/* For the first few MQI calls, we don't know if
 * an MCA is going to be a SVRCONN or not. The only calls
 * being intercepted by this particular module (PUT/GET)
 * will be done after that decision has been made, so by the
 * time we get here we'll be able to set the flag unconditionally and  
 * then continue. It's not until the decision is made that you can
 * check the userids. More sophisticated exits may like to do more
 * tracking in this function before eg unregistering calls so a regular
 * MCA isn't tracked.
 */
static void ClientCheck(struct mqstats *pmqstats,
                   PMQAXP pExitParms,
                   PMQAXC pExitContext)
{
  pmqstats->Environment = pExitContext->Environment;
  return;
}



all: apixstat 
apixstat :  apixstat.c Makefile
   /opt/SUNWspro/bin/cc -KPIC -mt -o $@ $@.c \
           -L/opt/mqm/lib -dy -G -lmqmzf -lmqm -lmqmcs -lmqmzse


all: apixstat apixstat_r
apixstat :  apixstat.c Makefile
   $(CC) -e MQStart -o $@   apixstat.c $(LDOPT) \
   -L/usr/mqm/lib  -lmqmzf -lmqm
apixstat_r : apixstat.c Makefile
   cc_r -D_REENTRANT  -e MQStart -o $@   apixstat.c \
-L/usr/mqm/lib -lmqmzf_r -lmqm_r


all: apixstat apixstat_r
LDOPT=-Wl,-rpath /opt/mqm/lib
apixstat :  apixstat.c Makefile
   $(CC) -fPIC -shared -e MQStart -o $@ apixstat.c \
   $(LDOPT) -L/opt/mqm/lib                         \
   -lmqmzf -lmqm -lmqmcs -lmqmzse -ldl -lc
apixstat_r : apixstat.c Makefile
   $(CC) -D_REENTRANT -D_POSIX_PTHREAD_SEMANTICS -fPIC  \
   $(LDOPT) -shared -e MQStart -o $@ apixstat.c -L/opt/mqm/lib \
-lmqmzf_r -lmqm_r -lmqmcs_r -lmqmzse -ldl -lc -lpthread


ApiExitLocal:
   Sequence=100
   Function=EntryPoint
   Module=/var/mqm/exits/apixstat
   Name=SampleApiExit


message <App:runmqsc      User:metaylor     Env:00000004 MsgPut:0 MsgGot:0 MsgBrw:0 BytesPut:0 BytesGot:0 BytesBrw: 0 MaxPut:0 MaxGot:0 >
message <App:ClientProxy  User:root         Env:00000002 MsgPut:2 MsgGot:0 MsgBrw:0 BytesPut:12 BytesGot:0 BytesBrw: 0 MaxPut:7 MaxGot:0>
message <App:ClientProxy  User:root         Env:00000002 MsgPut:27 MsgGot:0 MsgBrw:0 BytesPut:1857 BytesGot:0 BytesBrw: 0 MaxPut:81 MaxGot:0 >
message <App:ClientProxy  User:root         Env:00000002 MsgPut:1 MsgGot:0 MsgBrw:0 BytesPut:5 BytesGot:0 BytesBrw: 0 MaxPut:5 MaxGot:0>
message <App:amqsget      User:metaylor     Env:00000000 MsgPut:0 MsgGot:30 MsgBrw:0 BytesPut:0 BytesGot:1874 BytesBrw: 0 MaxPut:0 MaxGot:81 >
no more messages
Sample AMQSGET0 end

